@@ -2,7 +2,7 @@
 
 # MIT license
 #
-# Copyright (C) 2015-2019 by XESS Corp.
+# Copyright (C) 2015-2021 by Dave Vandenbout.
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -34,11 +34,10 @@ import sys
 import zipfile
 from builtins import str
 from collections import OrderedDict
-from copy import copy
+from copy import copy, deepcopy
 from pprint import pprint
 
 from affine import Affine
-from past.utils import old_div
 
 from .common import *
 from .pckg_info import __version__
@@ -60,41 +59,45 @@ PIN_LENGTH = 200
 PIN_SPACING = 100
 PIN_NUM_SIZE = 50  # Font size for pin numbers.
 PIN_NAME_SIZE = 50  # Font size for pin names.
-PIN_NAME_OFFSET = 40  # Separation between pin and pin name.
+PIN_NAME_OFFSET = 20  # Separation between pin and pin name.
 PIN_ORIENTATION = "left"
 PIN_STYLE = "line"
 SHOW_PIN_NUMBER = True  # Show pin numbers when True.
 SHOW_PIN_NAME = True  # Show pin names when True.
 SINGLE_PIN_SUFFIX = ""
-MULTI_PIN_SUFFIX = "*"
 PIN_SPACER_PREFIX = "*"
 
 # Settings for box drawn around pins in a unit.
-BOX_LINE_WIDTH = 12
+DEFAULT_BOX_LINE_WIDTH = 10
+
+# Mapping from understandable schematic symbol box fill-type name
+# to the fill-type indicator used in the KiCad part library.
+BOX_FILLS = {"no_fill": "N", "fg_fill": "F", "bg_fill": "f"}
+DEFAULT_BOX_FILL = "bg_fill"
 
 # Part reference.
-REF_SIZE = 60  # Font size.
-REF_Y_OFFSET = 250
+REF_SIZE = 50  # Font size.
+REF_Y_OFFSET = 0
 
 # Part number.
-PART_NUM_SIZE = 60  # Font size.
-PART_NUM_Y_OFFSET = 150
+PART_NUM_SIZE = 50  # Font size.
+PART_NUM_Y_OFFSET = -100
 
 # Part footprint
-PART_FOOTPRINT_SIZE = 60  # Font size.
-PART_FOOTPRINT_Y_OFFSET = 50
+PART_FOOTPRINT_SIZE = 50  # Font size.
+PART_FOOTPRINT_Y_OFFSET = -200
 
 # Part manufacturer number.
-PART_MPN_SIZE = 60  # Font size.
-PART_MPN_Y_OFFSET = -50
+PART_MPN_SIZE = 50  # Font size.
+PART_MPN_Y_OFFSET = -300
 
 # Part datasheet.
-PART_DATASHEET_SIZE = 60  # Font size.
-PART_DATASHEET_Y_OFFSET = -150
+PART_DATASHEET_SIZE = 50  # Font size.
+PART_DATASHEET_Y_OFFSET = -400
 
 # Part description.
-PART_DESC_SIZE = 60  # Font size.
-PART_DESC_Y_OFFSET = -250
+PART_DESC_SIZE = 50  # Font size.
+PART_DESC_Y_OFFSET = -500
 
 # Mapping from understandable pin orientation name to the orientation
 # indicator used in the KiCad part library. This mapping looks backward,
@@ -214,12 +217,9 @@ PIN_STYLES = {
 }
 PIN_STYLES = {scrubber.sub("", k).lower(): v for k, v in list(PIN_STYLES.items())}
 
-# Mapping from understandable box fill-type name to the fill-type
-# indicator used in the KiCad part library.
-FILLS = {"no_fill": "N", "fg_fill": "F", "bg_fill": "f"}
-
 # Format strings for various items in a KiCad part library.
 LIB_HEADER = "EESchema-LIBRARY Version 2.3\n"
+LIB_FOOTER = "#End Library\n"
 START_DEF = "DEF {name} {ref} 0 {pin_name_offset} {show_pin_number} {show_pin_name} {num_units} L N\n"
 END_DEF = "ENDDEF\n"
 REF_FIELD = 'F0 "{ref_prefix}" {x} {y} {font_size} H V {text_justification} CNN\n'
@@ -235,7 +235,7 @@ BOX = "S {x0} {y0} {x1} {y1} {unit_num} 1 {line_width} {fill}\n"
 PIN = "X {name} {num} {x} {y} {length} {orientation} {num_sz} {name_sz} {unit_num} 1 {pin_type} {pin_style}\n"
 
 
-def annotate_pins(unit_pins):
+def annotate_pins(unit_pins, annotation_style):
     """Annotate pin names to indicate special information."""
     for name, pins in unit_pins:
         # If there are multiple pins with the same name in a unit, then append a
@@ -245,8 +245,10 @@ def annotate_pins(unit_pins):
         # net connection in the schematic.)
         name_suffix = SINGLE_PIN_SUFFIX
         if len(pins) > 1:
-            # name_suffix = MULTI_PIN_SUFFIX
-            name_suffix = "[{}]".format(len(pins))
+            if annotation_style == "count":
+                name_suffix = "[{}]".format(len(pins))
+            elif annotation_style == "range":
+                name_suffix = "[{}:0]".format(len(pins) - 1)
         for pin in pins:
             pin.name += name_suffix
 
@@ -286,42 +288,51 @@ def pins_bbox(unit_pins):
     if len(unit_pins) == 0:
         return [[XO, YO], [XO, YO]]  # No pins, so no bounding box.
 
-    width = 0
-    for name, pins in unit_pins:
+    # Find the length of the longest pin name.
+    width = max((len(up[1][0].name) for up in unit_pins))
 
-        # Update the maximum observed width of a pin name. This is used later to
-        # size the width of the box surrounding the pin names for this unit.
-        width = max(width, len(pins[0].name) * PIN_NAME_SIZE)
-
-    # Add the separation space before and after the pin name.
-    width += PIN_LENGTH + 2 * PIN_NAME_OFFSET
+    # Compute bbox width adding the separation space before and after the pin name.
+    # (The space after the pin name keeps pin names on opposing sides from colliding.)
+    width = width * PIN_NAME_SIZE + PIN_LENGTH + 2 * PIN_NAME_OFFSET
 
     # Make bounding box an integer number of pin spaces so pin connections are always on the grid.
-    width = math.ceil(old_div(float(width), PIN_SPACING)) * PIN_SPACING
+    width = math.ceil(float(width) / PIN_SPACING) * PIN_SPACING
 
     # Compute the height of the column of pins.
     height = count_pin_slots(unit_pins) * PIN_SPACING
 
+    # Return the bounding box including a spacer on each end of the column.
     return [[XO, YO + PIN_SPACING], [XO + width, YO - height]]
 
 
-def balance_bboxes(bboxes):
-    """Make the symbol more balanced by adjusting the bounding boxes of the pins on each side."""
-    X = 0
-    Y = 1
+def find_bbox_bbox(*bboxes):
+    """Find the bounding box for a set of bounding boxes."""
 
-    def find_bbox_bbox(*bboxes):
-        """Find the bounding box for a set of bounding boxes."""
-        bb = [[0, 0], [0, 0]]
-        for bbox in bboxes:
-            bb[0][X] = min(bb[0][X], bbox[0][X])
-            bb[1][X] = max(bb[1][X], bbox[1][X])
-            bb[0][Y] = max(bb[0][Y], bbox[0][Y])
-            bb[1][Y] = min(bb[1][Y], bbox[1][Y])
-        return bb
+    X, Y = 0, 1
+
+    bb = [[0, 0], [0, 0]]
+    for bbox in bboxes:
+        bb[0][X] = min(bb[0][X], bbox[0][X])
+        bb[1][X] = max(bb[1][X], bbox[1][X])
+        bb[0][Y] = max(bb[0][Y], bbox[0][Y])
+        bb[1][Y] = min(bb[1][Y], bbox[1][Y])
+    return bb
+
+
+def balance_bboxes(bboxes):
+    """
+    Make the symbol more balanced by adjusting the bounding boxes of the pins on each side.
+    """
+
+    X, Y = 0, 1
+
+    non_empty_sides = []
+    for side, bbox in bboxes.items():
+        if bbox[0][X] != bbox[1][X] and bbox[0][Y] != bbox[1][Y]:
+            non_empty_sides.append(side)
 
     # Determine the number of sides of the symbol with pins.
-    num_sides = len(bboxes)
+    num_sides = len(non_empty_sides)
 
     if num_sides == 4:
         # If the symbol has pins on all four sides, then check to see if there
@@ -332,39 +343,39 @@ def balance_bboxes(bboxes):
         lr_hgt = abs(lr_bbox[0][Y] - lr_bbox[1][Y])
         tb_bbox = find_bbox_bbox(bboxes["top"], bboxes["bottom"])
         tb_hgt = abs(tb_bbox[0][Y] - tb_bbox[1][Y])
-        if 0.75 <= float(lr_hgt) / float(tb_hgt) <= 1 / 0.75:
+        if 0.75 * tb_hgt <= lr_hgt <= 1.33 * tb_hgt:
             bal_bbox = find_bbox_bbox(*list(bboxes.values()))
             for side in bboxes:
-                bboxes[side] = copy(bal_bbox)
+                bboxes[side] = deepcopy(bal_bbox)
         else:
-            bboxes["left"] = copy(lr_bbox)
-            bboxes["right"] = copy(lr_bbox)
-            bboxes["top"] = copy(tb_bbox)
-            bboxes["bottom"] = copy(tb_bbox)
+            bboxes["left"] = deepcopy(lr_bbox)
+            bboxes["right"] = deepcopy(lr_bbox)
+            bboxes["top"] = deepcopy(tb_bbox)
+            bboxes["bottom"] = deepcopy(tb_bbox)
     elif num_sides == 3:
-        # If the symbol only has pins on threee sides, then equalize the
+        # If the symbol only has pins on three sides, then equalize the
         # bounding boxes for the pins on opposite sides and leave the
         # bounding box on the other side unchanged.
-        if "left" not in bboxes or "right" not in bboxes:
+        if "top" in non_empty_sides and "bottom" in non_empty_sides:
             # Top & bottom side pins, but the left or right side is empty.
             bal_bbox = find_bbox_bbox(bboxes["top"], bboxes["bottom"])
-            bboxes["top"] = copy(bal_bbox)
-            bboxes["bottom"] = copy(bal_bbox)
-        elif "top" not in bboxes or "bottom" not in bboxes:
+            bboxes["top"] = deepcopy(bal_bbox)
+            bboxes["bottom"] = deepcopy(bal_bbox)
+        elif "left" in non_empty_sides and "right" in non_empty_sides:
             # Left & right side pins, but the top or bottom side is empty.
             bal_bbox = find_bbox_bbox(bboxes["left"], bboxes["right"])
-            bboxes["left"] = copy(bal_bbox)
-            bboxes["right"] = copy(bal_bbox)
+            bboxes["left"] = deepcopy(bal_bbox)
+            bboxes["right"] = deepcopy(bal_bbox)
     elif num_sides == 2:
         # If the symbol only has pins on two opposing sides, then equalize the
         # height of the bounding boxes for each side. Leave the width unchanged.
-        if "left" in bboxes and "right" in bboxes:
+        if "left" in non_empty_sides and "right" in non_empty_sides:
             bal_bbox = find_bbox_bbox(bboxes["left"], bboxes["right"])
             bboxes["left"][0][Y] = bal_bbox[0][Y]
             bboxes["left"][1][Y] = bal_bbox[1][Y]
             bboxes["right"][0][Y] = bal_bbox[0][Y]
             bboxes["right"][1][Y] = bal_bbox[1][Y]
-        elif "top" in bboxes and "bottom" in bboxes:
+        elif "top" in non_empty_sides and "bottom" in non_empty_sides:
             bal_bbox = find_bbox_bbox(bboxes["top"], bboxes["bottom"])
             bboxes["top"][0][Y] = bal_bbox[0][Y]
             bboxes["top"][1][Y] = bal_bbox[1][Y]
@@ -372,7 +383,23 @@ def balance_bboxes(bboxes):
             bboxes["bottom"][1][Y] = bal_bbox[1][Y]
 
 
-def draw_pins(unit_num, unit_pins, bbox, transform, fuzzy_match):
+def calc_scrunch(bboxes):
+    """Return the "scrunch" for compressing the left/right sides under the top/bottom."""
+
+    X, Y = 0, 1
+
+    # Calculate how much the left/right sides can be "scrunched in" so they
+    # reside underneath the top/bottom rows.
+    lr_bbox = find_bbox_bbox(bboxes["left"], bboxes["right"])
+    tb_bbox = find_bbox_bbox(bboxes["top"], bboxes["bottom"])
+    tb_width = abs(tb_bbox[0][Y] - tb_bbox[1][Y])
+    lr_width = abs(lr_bbox[0][X] - lr_bbox[1][X]) - PIN_LENGTH
+    scrunch = min(tb_width / 2, lr_width)
+
+    return scrunch
+
+
+def draw_pins(unit_num, unit_pins, bbox, transform, side, push, fuzzy_match):
     """Draw a column of pins rotated/translated by the transform matrix."""
 
     # String to add pin definitions to.
@@ -384,10 +411,11 @@ def draw_pins(unit_num, unit_pins, bbox, transform, fuzzy_match):
     Y = 1  # Index for Y coordinate.
     pins_bb = pins_bbox(unit_pins)
     height_offset = abs(bbox[0][Y] - bbox[1][Y]) - abs(pins_bb[0][Y] - pins_bb[1][Y])
-    height_offset /= 2
-    height_offset -= (
-        height_offset % PIN_SPACING
-    )  # Keep everything on the PIN_SPACING grid.
+    push = min(max(0.0, push), 1.0)
+    if side in ("right", "top"):
+        push = 1.0 - push
+    height_offset *= push
+    height_offset -= height_offset % PIN_SPACING  # Keep stuff on the PIN_SPACING grid.
 
     # Start drawing pins from the origin.
     x = XO
@@ -417,9 +445,9 @@ def draw_pins(unit_num, unit_pins, bbox, transform, fuzzy_match):
             pin_style = "N" + pin_style
 
         # Create all the pins with a particular name. If there are more than one,
-        # they are laid on top of each other and only the first is visible.
-        num_size = PIN_NUM_SIZE  # First pin will be visible.
-        for pin in pins:
+        # pin numbers are hidden, and everything after the first are hidden.
+        # num_size = PIN_NUM_SIZE if len(pins) == 1 else 0
+        for index, pin in enumerate(pins):
 
             pin_num = str(pin.num)
 
@@ -435,15 +463,23 @@ def draw_pins(unit_num, unit_pins, bbox, transform, fuzzy_match):
                 y=int(draw_y),
                 length=PIN_LENGTH,
                 orientation=pin_side,
-                num_sz=num_size,
+                num_sz=PIN_NUM_SIZE,
+                # num_sz=num_size,
                 name_sz=PIN_NAME_SIZE,
                 unit_num=unit_num,
                 pin_type=pin_type,
                 pin_style=pin_style,
             )
 
-            # Turn off visibility after the first pin.
-            num_size = 0
+            # Make tweaks to subsequent bundled pins:
+            # make them invisible
+            pin_style = "N" + pin_style.lstrip("N")
+            # power pins become passive pins
+            if pin_type in "wW":
+                pin_type = "P"
+            # NC pins should be shifted off-grid into the symbol to avoid shorts
+            if pin_type == "N":
+                (draw_x, draw_y) = transform * (x + index, y + 1)
 
         # Move to the next pin placement location on this unit.
         y -= PIN_SPACING
@@ -463,27 +499,32 @@ def zero_pad_nums(s):
         return s  # The input is probably not a string, so just return it unchanged.
 
 
+def str_to_num_alpha_tuple(s):
+    # Split a string of alphas and digits into a tuple of alpha/digit strings.
+    try:
+        seq = re.split(r"(?<=\D)(?=\d)|(?<=\d)(?=\D)", s)
+    except ValueError:
+        return (zero_pad_nums(s),)
+    return tuple(zero_pad_nums(_) for _ in seq)
+
+
 def num_key(pin):
     """Generate a key from a pin's number so they are sorted by position on the package."""
-
-    # Pad all numeric strings in the pin name with leading 0's.
-    # Thus, 'A10' and 'A2' will become 'A00010' and 'A00002' and A2 will
-    # appear before A10 in a list.
-    return zero_pad_nums(pin[1][0].num)
+    return str_to_num_alpha_tuple(pin[1][0].num) + str_to_num_alpha_tuple(
+        pin[1][0].name
+    )
 
 
 def name_key(pin):
-    """Generate a key from a pin's name so they are sorted more logically."""
-
-    # Pad all numeric strings in the pin name with leading 0's.
-    # Thus, 'adc10' and 'adc2' will become 'adc00010' and 'adc00002' and adc2 will
-    # appear before adc10 in a list.
-    return zero_pad_nums(pin[1][0].name)
+    "Generate a key from a pin's name so they are sorted more logically." ""
+    return str_to_num_alpha_tuple(pin[1][0].name) + str_to_num_alpha_tuple(
+        pin[1][0].num
+    )
 
 
 def row_key(pin):
     """Generate a key from the order the pins were entered into the CSV file."""
-    return pin[1][0].index
+    return (pin[1][0].index,)
 
 
 def draw_symbol(
@@ -498,6 +539,11 @@ def draw_symbol(
     reverse,
     fuzzy_match,
     fill,
+    box_line_width,
+    push,
+    annotation_style,
+    center_symbol,
+    scrunch,
 ):
     """Add a symbol for a part to the library."""
 
@@ -511,16 +557,24 @@ def draw_symbol(
         num_units=len(pin_data),
     )
 
-    # Determine if there are pins across the top of the symbol.
-    # If so, right-justify the reference, part number, etc. so they don't
-    # run into the top pins. If not, stick with left-justification.
-    text_justification = "L"
-    horiz_offset = PIN_LENGTH
-    for unit in list(pin_data.values()):
-        if "top" in list(unit.keys()):
-            text_justification = "R"
-            horiz_offset = PIN_LENGTH - 50
-            break
+    if center_symbol:
+        # Origin is in the center of the symbol.
+        text_justification = "C"
+        horiz_offset = 0
+        vert_offset = 50
+    else:
+        # Origin is on the top-most pin on the left side.
+        # Determine if there are pins across the top of the symbol.
+        # If so, right-justify the reference, part number, etc. so they don't
+        # run into the top pins. If not, stick with left-justification.
+        text_justification = "L"
+        horiz_offset = PIN_LENGTH
+        vert_offset = 250
+        for unit in list(pin_data.values()):
+            if "top" in list(unit.keys()):
+                text_justification = "R"
+                horiz_offset = PIN_LENGTH - 50
+                break
 
     # Create the field that stores the part reference.
     if not part_ref_prefix:
@@ -528,7 +582,7 @@ def draw_symbol(
     part_defn += REF_FIELD.format(
         ref_prefix=part_ref_prefix,
         x=XO + horiz_offset,
-        y=YO + REF_Y_OFFSET,
+        y=YO + REF_Y_OFFSET + vert_offset,
         text_justification=text_justification,
         font_size=REF_SIZE,
     )
@@ -539,7 +593,7 @@ def draw_symbol(
     part_defn += PARTNUM_FIELD.format(
         num=part_num,
         x=XO + horiz_offset,
-        y=YO + PART_NUM_Y_OFFSET,
+        y=YO + PART_NUM_Y_OFFSET + vert_offset,
         text_justification=text_justification,
         font_size=PART_NUM_SIZE,
     )
@@ -550,7 +604,7 @@ def draw_symbol(
     part_defn += FOOTPRINT_FIELD.format(
         footprint=part_footprint,
         x=XO + horiz_offset,
-        y=YO + PART_FOOTPRINT_Y_OFFSET,
+        y=YO + PART_FOOTPRINT_Y_OFFSET + vert_offset,
         text_justification=text_justification,
         font_size=PART_FOOTPRINT_SIZE,
     )
@@ -561,7 +615,7 @@ def draw_symbol(
     part_defn += DATASHEET_FIELD.format(
         datasheet=part_datasheet,
         x=XO + horiz_offset,
-        y=YO + PART_DATASHEET_Y_OFFSET,
+        y=YO + PART_DATASHEET_Y_OFFSET + vert_offset,
         text_justification=text_justification,
         font_size=PART_DATASHEET_SIZE,
     )
@@ -571,7 +625,7 @@ def draw_symbol(
         part_defn += MPN_FIELD.format(
             manf_num=part_manf_num,
             x=XO + horiz_offset,
-            y=YO + PART_MPN_Y_OFFSET,
+            y=YO + PART_MPN_Y_OFFSET + vert_offset,
             text_justification=text_justification,
             font_size=PART_MPN_SIZE,
         )
@@ -581,7 +635,7 @@ def draw_symbol(
         part_defn += DESC_FIELD.format(
             desc=part_desc,
             x=XO + horiz_offset,
-            y=YO + PART_DESC_Y_OFFSET,
+            y=YO + PART_DESC_Y_OFFSET + vert_offset,
             text_justification=text_justification,
             font_size=PART_DESC_SIZE,
         )
@@ -615,16 +669,18 @@ def draw_symbol(
 
         # Annotate the pins for each side of the symbol.
         for side_pins in list(unit.values()):
-            annotate_pins(list(side_pins.items()))
+            annotate_pins(list(side_pins.items()), annotation_style)
 
         # Determine the actual bounding box for each side.
-        bbox = {}
         for side, side_pins in list(unit.items()):
             bbox[side] = pins_bbox(list(side_pins.items()))
 
         # Adjust the sizes of the bboxes to make the unit look more symmetrical.
         balance_bboxes(bbox)
 
+        # Calculate the amount to move the left/right sides to move them under the top/bottom rows.
+        scrunch_offset = calc_scrunch(bbox) if scrunch else 0
+
         # Determine some important points for each side of pins.
         for side in unit:
             #
@@ -670,17 +726,12 @@ def draw_symbol(
         #        +--------+
         #
 
-        # Create zero-sized bounding boxes for any sides of the unit without pins.
-        # This makes it simpler to do the width/height calculation that follows.
-        for side in all_sides:
-            if side not in bbox:
-                bbox[side] = [(XO, YO), (XO, YO)]
-
         # This is the width and height of the box in the middle of the pins on each side.
         box_width = max(
             abs(bbox["top"][0][Y] - bbox["top"][1][Y]),
             abs(bbox["bottom"][0][Y] - bbox["bottom"][1][Y]),
         )
+        box_width -= 2 * scrunch_offset
         box_height = max(
             abs(bbox["left"][0][Y] - bbox["left"][1][Y]),
             abs(bbox["right"][0][Y] - bbox["right"][1][Y]),
@@ -703,9 +754,10 @@ def draw_symbol(
             elif side == "bottom":
                 # Translate AL to AB
                 translate_y -= box_height
+                translate_x -= scrunch_offset
             elif side == "top":
                 # Translate AL to AT
-                translate_x += box_width
+                translate_x += box_width + scrunch_offset
             # Create the complete transformation matrix = rotation followed by translation.
             transform[side] = (
                 Affine.translation(translate_x, translate_y) * transform[side]
@@ -713,8 +765,33 @@ def draw_symbol(
             # Also translate the point on each side that defines the box around the symbol.
             box_pt[side] = transform[side] * box_pt[side]
 
+        if center_symbol:
+            # If centering, compute the translation to move the symbol box center to the origin.
+            bbox_translate_x = -(int(box_pt["right"][X]) + int(box_pt["left"][X])) / 2
+            bbox_translate_y = -(int(box_pt["bottom"][Y]) + int(box_pt["top"][Y])) / 2
+
+            # Force the translation to a multiple of the pin spacing so the pins don't
+            # go off grid.
+            bbox_translate_x = round(bbox_translate_x / PIN_SPACING) * PIN_SPACING
+            bbox_translate_y = round(bbox_translate_y / PIN_SPACING) * PIN_SPACING
+
+            # Add the translation to all the affine transforms of the sides of pins.
+            for side in all_sides:
+                transform[side] = (
+                    Affine.translation(bbox_translate_x, bbox_translate_y)
+                    * transform[side]
+                )
+
+            # Also translate the point on each side that defines the box around the symbol.
+            for side in all_sides:
+                box_pt[side] = (
+                    Affine.translation(bbox_translate_x, bbox_translate_y)
+                    * box_pt[side]
+                )
+
         # Draw the transformed pins for each side of the symbol.
-        for side, side_pins in list(unit.items()):
+        for side in all_sides:
+            side_pins = unit[side]
             # If the pins are ordered by their row in the spreadsheet or by their name,
             # then reverse their order on the right and top sides so they go from top-to-bottom
             # on the right side and left-to-right on the top side instead of the opposite
@@ -728,18 +805,24 @@ def draw_symbol(
             )
             # Draw the transformed pins for this side of the symbol.
             part_defn += draw_pins(
-                unit_num, sorted_side_pins, bbox[side], transform[side], fuzzy_match
+                unit_num,
+                sorted_side_pins,
+                bbox[side],
+                transform[side],
+                side,
+                push,
+                fuzzy_match,
             )
 
-            # Create the box around the unit's pins.
+        # Create the box around the unit's pins.
         part_defn += BOX.format(
             x0=int(box_pt["left"][X]),
             y0=int(box_pt["top"][Y]),
             x1=int(box_pt["right"][X]),
             y1=int(box_pt["bottom"][Y]),
             unit_num=unit_num,
-            line_width=BOX_LINE_WIDTH,
-            fill=FILLS[fill],
+            line_width=box_line_width,
+            fill=BOX_FILLS[fill],
         )
 
     # Close the section that holds the part's units.
@@ -752,23 +835,25 @@ def draw_symbol(
     return part_defn
 
 
-def is_pwr(pin, fuzzy_match):
-    """Return true if this is a power input pin."""
+def can_bundle(pin, fuzzy_match):
+    """Return true if the pin type can be bundled.
+    Currently support bundling power input, power output, and NC pins.
+    """
     return (
         find_closest_match(name=pin.type, name_dict=PIN_TYPES, fuzzy_match=fuzzy_match)
-        == "W"
+        in "wWN"
     )
 
 
 def do_bundling(pin_data, bundle, fuzzy_match):
-    """Handle bundling for power pins. Unbundle everything else."""
+    """Handle bundling. Unbundle everything else."""
     for unit in list(pin_data.values()):
         for side in list(unit.values()):
             for name, pins in list(side.items()):
                 if len(pins) > 1:
                     for index, p in enumerate(pins):
-                        if is_pwr(p, fuzzy_match) and bundle:
-                            side[p.name + "_pwr"].append(p)
+                        if can_bundle(p, fuzzy_match) and bundle:
+                            side[p.name + "_bundle"].append(p)
                         else:
                             side[p.name + "_" + str(index)].append(p)
                     del side[name]
@@ -794,11 +879,16 @@ def kipart(
     part_data_file_type,
     parts_lib,
     fill,
+    box_line_width,
+    push,
     allow_overwrite=False,
     sort_type="name",
     reverse=False,
     fuzzy_match=False,
     bundle=False,
+    annotation_style="count",
+    center_symbol=False,
+    scrunch=False,
     debug_level=0,
 ):
     """Read part pin data from a CSV/text/Excel file and write or append it to a library file."""
@@ -837,6 +927,11 @@ def kipart(
             reverse=reverse,
             fuzzy_match=fuzzy_match,
             fill=fill,
+            box_line_width=box_line_width,
+            push=push,
+            annotation_style=annotation_style,
+            center_symbol=center_symbol,
+            scrunch=scrunch,
         )
 
 
@@ -860,11 +955,11 @@ def read_lib_file(lib_file):
 
 def write_lib_file(parts_lib, lib_file):
     print("Writing", lib_file, len(parts_lib))
-    LIB_HEADER = "EESchema-LIBRARY Version 2.3\n"
     with open(lib_file, "w") as lib_fp:
         lib_fp.write(LIB_HEADER)
         for part_def in parts_lib.values():
             lib_fp.write(part_def)
+        lib_fp.write(LIB_FOOTER)
 
 
 def call_kipart(args, part_reader, part_data_file, file_name, file_type, parts_lib):
@@ -876,11 +971,16 @@ def call_kipart(args, part_reader, part_data_file, file_name, file_type, parts_l
         part_data_file_type=file_type,
         parts_lib=parts_lib,
         fill=args.fill,
+        box_line_width=args.box_line_width,
+        push=args.push,
         allow_overwrite=args.overwrite,
         sort_type=args.sort,
         reverse=args.reverse,
         fuzzy_match=args.fuzzy_match,
         bundle=args.bundle,
+        annotation_style=args.annotation_style,
+        center_symbol=args.center,
+        scrunch=args.scrunch,
         debug_level=args.debug,
     )
 
@@ -935,6 +1035,26 @@ def main():
         default="left",
         help="Which side to place the pins by default.",
     )
+    parser.add_argument(
+        "--fill",
+        nargs="?",
+        type=lambda s: unicode(s).lower(),
+        choices=BOX_FILLS.keys(),
+        default=DEFAULT_BOX_FILL,
+        help="Select fill style for schematic symbol boxes.",
+    )
+    parser.add_argument(
+        "--box_line_width",
+        type=int,
+        default=DEFAULT_BOX_LINE_WIDTH,
+        help="Set line width of the schematic symbol box.",
+    )
+    parser.add_argument(
+        "--push",
+        type=float,
+        default=0.5,
+        help="Push pins left/up (0.0), center (0.5), or right/down(1.0) on the sides of the schematic symbol box.",
+    )
     parser.add_argument(
         "-o",
         "--output",
@@ -955,6 +1075,25 @@ def main():
         action="store_true",
         help="Bundle multiple, identically-named power and ground pins each into a single schematic pin.",
     )
+    parser.add_argument(
+        "--annotation_style",
+        nargs="?",
+        type=lambda s: unicode(s).lower(),
+        choices=["none", "count", "range"],
+        default="count",
+        help="Selects suffix appended to bundled pin names: none (), count ([n]), range ([n:0]).",
+    )
+    parser.add_argument(
+        "-c",
+        "--center",
+        action="store_true",
+        help="Place symbol origin in the center of the symbol.",
+    )
+    parser.add_argument(
+        "--scrunch",
+        action="store_true",
+        help="Compress pins of left/right columns underneath top/bottom rows of pins.",
+    )
     parser.add_argument(
         "-a",
         "--append",
@@ -977,14 +1116,6 @@ def main():
         metavar="LEVEL",
         help="Print debugging info. (Larger LEVEL means more info.)",
     )
-    parser.add_argument(
-        "--fill",
-        nargs="?",
-        type=lambda s: unicode(s).lower(),
-        choices=["no_fill", "fg_fill", "bg_fill"],
-        default="no_fill",
-        help="Select fill/no-fill for schematic symbol boxes.",
-    )
 
     args = parser.parse_args()
 
